/* The zlib/libpng License

Copyright (c) 2005 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
   not claim that you wrote the original software. If you use this
   software in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
   distribution.
*/

#include "sndfile.hpp"
#include "fft.hpp"
#include <sstream>
#include <iomanip>
#include <unistd.h>
using namespace doj;
using namespace std;

#include "common.cpp"

void help()
{
  cerr << "usage: sample2powerspectrum [-2..-9] [-b] [-d] [-h height] [-l] [-o offset] [-s fontsize] [-t text] [-w width] <filename>\n"
       << "  -2..-9 print only first nth of power spectrum\n"
       << "  -b draw box around image\n"
       << "  -d draw frequencies and lines\n"
       << "  -h height in cm\n"
       << "  -l plot intensities logarithmically\n"
       << "  -o offset in cm from lower left corner\n"
       << "  -s fontsize in pt\n"
       << "  -t print text in lower left corner. if text is a single space character the filename and samplerate are printed\n"
       << "  -w width in cm\n"
       << "version " << VERSION << "\n"
       << "convered by the zlib/libpng license\n"
       << "(c) by Dirk Jagdmann <doj@cubic.org>\n"
       << "http://llg.cubic.org/tools/sample2ps\n"
    ;

  exit(1);
}

int main(int argc, char **argv)
{
  // parse command line
  double width=800, height=-1, offset=0;
  bool drawbox=false, dolog=false, drawlines=false;
  string text;
  int c, scale=1;
  while((c=getopt(argc, argv, "123456789bdh:lo:s:t:w:")) >= 0)
    switch(c)
      {
      default: help();

      case '1': scale=1; break;
      case '2': scale=2; break;
      case '3': scale=3; break;
      case '4': scale=4; break;
      case '5': scale=5; break;
      case '6': scale=6; break;
      case '7': scale=7; break;
      case '8': scale=8; break;
      case '9': scale=9; break;

      case 'b': drawbox=true; break;

      case 'd': drawlines=true; break;

      case 'h':
	height=strtof(optarg,0)/2.54*72.0;
	if(height<=0.0)
	  help();
	break;

      case 'l': dolog=true; break;

      case 'o': offset=strtof(optarg,0)/2.54*72.0; break;

      case 's':
	fontsize=strtof(optarg,0);
	if(fontsize<=0.0)
	  help();
	break;

      case 't': text=optarg; break;

      case 'w':
	width=strtof(optarg,0)/2.54*72.0;
	if(width<=0.0)
	  help();
	break;
      }

  if(height<0.0)
    height=width/1.6;

  if(optind==argc)
    help();

  // load sample
  const string filename=argv[optind];
  int samplerate, channels;
  unsigned size;
  double *buf=0;
  buf=sndfile_read(filename, samplerate, channels, size, buf);
  if(!buf)
    return 1;
  if(channels!=1)
    {
      cerr << "channels!=1\n";
      return 1;
    }
  if(!size)
    {
      cerr << "size==0\n";
      return 1;
    }

  // calculate power spectrum
  typedef std::vector<double> v_t;
  v_t ps=powerspectrum(buf, size, doj::blackman_window);
  double max=0;
  for(v_t::iterator i=ps.begin(); i!=ps.end(); ++i)
    {
      if(dolog)
	*i=log(*i+1);
      if(*i>max)
	max=*i;
    }
  if(max!=0.0)
    {
      max=1.0/max;
      for(v_t::iterator i=ps.begin(); i!=ps.end(); ++i)
	*i *= max;
    }

  // setup float format
  cout << fixed << setprecision(2);

  cout << "%!PS-Adobe-3.0 EPSF-3.0\n"
       << "%%BoundingBox: " << ((offset!=0.0)?static_cast<int>(offset-1.0):0) << " "
                            << ((offset!=0.0)?static_cast<int>(offset-1.0):0) << " "
                            << static_cast<int>(width+offset+2.0) << " "
                            << static_cast<int>(height+offset+2.0) << "\n"
       << "%%Title: " << argv[optind] << "\n"
       << "%%Creator: sample2powerspectrum http://llg.cubic.org/tools/sample2ps\n"
       << "gsave\n"
       << "1 setlinecap 1 setlinejoin\n"

    // dashed lines
       << "% see Postscript Tutorial and Cookbook\n"
       << "/pathlength {\n"
       << "    flattenpath\n"
       << "    /dist 0 def\n"
       << "    {\n"
       << "	/yfirst exch def\n"
       << "	/xfirst exch def\n"
       << "	/ymoveto yfirst def\n"
       << "	/xmoveto xfirst def\n"
       << "    }\n"
       << "    {\n"
       << "	/ynext exch def\n"
       << "	/xnext exch def\n"
       << "	/dist dist ynext yfirst sub dup mul xnext xfirst sub dup mul add sqrt add def\n"
       << "	/yfirst ynext def\n"
       << "	/xfirst xnext def\n"
       << "    }\n"
       << "    { }\n"
       << "    {\n"
       << "	/ynext ymoveto def\n"
       << "	/xnext xmoveto def\n"
       << "	/dist dist ynext yfirst sub dup mul xnext xfirst sub dup mul add sqrt add def\n"
       << "	/yfirst ynext def\n"
       << "	/xfirst xnext def\n"
       << "    }\n"
       << "    pathforall\n"
       << "    dist\n"
       << "} def\n"
       << "/centerdash {\n"
       << "    /pattern exch  def\n"
       << "    /pathlen pathlength def\n"
       << "    /patternlength 0 def\n"
       << "    pattern { patternlength add /patternlength exch def } forall\n"
       << "    pattern length 2 mod 0 ne { /patternlength patternlength 2 mul def } if\n"
       << "    /first pattern 0 get def\n"
       << "    /last patternlength first sub def\n"
       << "    /n pathlen last sub cvi patternlength cvi idiv def\n"
       << "    /endpart pathlen patternlength n mul sub last sub 2 div def\n"
       << "    /offset first endpart sub def\n"
       << "    pattern offset setdash\n"
       << "} def\n"
       << "% newpath 0 0 moveto 800 500 lineto [20] centerdash stroke\n"
       << "% newpath 0 0 moveto 800 500 lineto [20 10] centerdash stroke\n"
       << "% newpath 0 0 moveto 800 500 lineto [20 10 3] centerdash stroke\n"

    // outer frame
       << "/fr { "
       << "  newpath "
       << "  0 0 moveto "
       << "  0 " << height << " lineto "
       << width << " " << height << " lineto "
       << width << " 0 lineto "
       << "  closepath "
       << "} def\n"
    ;

  if(offset!=0.0)
    cout << offset << " " << offset << " translate\n";

  // draw power spectrum
  {
    cout << "newpath 0 0 moveto\n";
    size=ps.size()/scale;
    for(unsigned i=1; i<size; ++i)
      {
	const double th=0.01;
#define DRAW(y) cout << width*i/size << " " << height*y << " lineto\n"
	if(ps[i]<th)
	  {
	    // search for a new offset
	    DRAW(ps[i]);
	    DRAW(0);
	    while(ps[i]<th && i<size) ++i;
	    --i;
	    DRAW(0);
	  }
	else
	  DRAW(ps[i]);
#undef DRAW
      }
    cout << "0.25 setlinewidth stroke\n";
  }

  setupFont("Times-Roman");

  // draw lines
  if(drawlines)
    {
      cout << "gsave [5] centerdash\n";
      setcol(0.7);
      int ls=static_cast<int>(pt2cm(width)/2.0);
      if(ls&1) ++ls;		// make ls even
      for(int i=1; i<=ls; ++i)
	{
	  const double x=width*i/ls;
	  cout << "newpath " << x << " 0 moveto " << x << " " << height << " lineto stroke\n";

	  ostringstream os; os << static_cast<int>(static_cast<double>(i)/ls*samplerate/2.0/scale);
	  print(os.str(), x, height-fontsize*((i==ls)?2:1), rgba(0,0,0), (i==ls)?Right:Center) << endl;
	}
      cout << "grestore\n";
    }

  // draw text
  if(text.size())
    {
      string t;
      if(text==" ")
	{
	  ostringstream os;
	  os << filename << " " << samplerate << "Hz";
	  t=os.str();
	}
      else
	t=text;

      print(t, 0.25*fontsize, 0.25*fontsize);
    }

  // draw box
  if(drawbox)
    setcol(0.0) << "fr 1 setlinewidth stroke\n";

  cout << "grestore\n";
  cout << "showpage\n";
  return 0;
}
