/* The zlib/libpng License

Copyright (c) 2005 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
   not claim that you wrote the original software. If you use this
   software in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
   distribution.
*/

#include "sndfile.hpp"
#include <sstream>
#include <iomanip>
#include <string>
#include <unistd.h>
using namespace doj;
using namespace std;

#include "common.cpp"

void help()
{
  cerr << "usage: sample2ps [-2..-9] [-a scale] [-b] [-h height] [-l size] [-o offset] [-s fontsize] [-t text] [-w width] [-z] <filename>\n"
       << "  -2..-9 scale samples\n"
       << "  -a scale by arbitrary value. scale>0\n"
       << "  -b draw box around image\n"
       << "  -h height in cm. height>0.0\n"
       << "  -l strokesize. strokesize>0.0\n"
       << "  -o offset in cm from lower left corner\n"
       << "  -s fontsize in pt. fontsize>0.0\n"
       << "  -t print text in lower left corner. if text is a single space character the filename and samplerate are printed\n"
       << "  -w width in cm. width>0.0\n"
       << "  -z draw timelines\n"
       << "version " << VERSION << "\n"
       << "convered by the zlib/libpng license\n"
       << "(c) by Dirk Jagdmann <doj@cubic.org>\n"
       << "http://llg.cubic.org/tools/sample2ps\n"
    ;

  exit(1);
}

int main(int argc, char **argv)
{
  float width=800, height=-1, offset=0, strokesize=0.25;
  bool drawbox=false, drawtime=false;
  string text;
  int c, scale=1;
  while((c=getopt(argc, argv, "123456789a:bh:l:o:s:t:w:z")) >= 0)
    switch(c)
      {
      default: help();

      case '1': scale=1; break;
      case '2': scale=2; break;
      case '3': scale=3; break;
      case '4': scale=4; break;
      case '5': scale=5; break;
      case '6': scale=6; break;
      case '7': scale=7; break;
      case '8': scale=8; break;
      case '9': scale=9; break;

      case 'a':
	scale=strtol(optarg,0,0);
	if(scale<1)
	  help();
	break;

      case 'b': drawbox=true; break;

      case 'h':
	height=strtof(optarg,0)/2.54*72.0;
	if(height<=0.0)
	  help();
	break;

      case 'l':
	strokesize=strtof(optarg,0);
	if(strokesize<=0.0)
	  help();
	break;

      case 'o': offset=strtof(optarg,0)/2.54*72.0; break;

      case 's':
	fontsize=strtof(optarg,0);
	if(fontsize<=0.0)
	  help();
	break;

      case 't': text=optarg; break;

      case 'w':
	width=strtof(optarg,0)/2.54*72.0;
	if(width<=0.0)
	  help();
	break;

      case 'z': drawtime=true; break;
      }

  if(height<0.0)
    height=width/1.6;

  if(optind==argc)
    help();

  const string filename=argv[optind];

  int samplerate, channels;
  unsigned size;
  float *buf=0;
  buf=sndfile_read(filename, samplerate, channels, size, buf);
  if(!buf)
    return 1;
  if(channels<1)
    {
      cerr << "channels<1\n";
      return 1;
    }
  if(!size)
    {
      cerr << "size==0\n";
      return 1;
    }

  const int frames=(size/channels)/scale;
  const float step=width/frames;
  const float amp=height/channels*.5;

  width+=strokesize*2.0;
  height+=strokesize*2.0;

  // setup float format
  cout << fixed << setprecision(2);

  cout << "%!PS-Adobe-3.0 EPSF-3.0\n"
       << "%%BoundingBox: " << ((offset!=0.0)?static_cast<int>(offset-1.0):0) << " "
                            << ((offset!=0.0)?static_cast<int>(offset-1.0):0) << " "
                            << static_cast<int>(width+offset+2.0) << " "
                            << static_cast<int>(height+offset+2.0) << "\n"
       << "%%Title: " << argv[optind] << "\n"
       << "%%Creator: sample2ps http://llg.cubic.org/tools/sample2ps\n"
       << "gsave\n"
       << "1 setlinecap 1 setlinejoin\n"

    // outer frame
       << "/fr { "
       << "  newpath "
       << "  0 0 moveto "
       << "  0 " << height << " lineto "
       << width << " " << height << " lineto "
       << width << " 0 lineto "
       << "  closepath "
       << "} def\n"
    ;

  setupFont("Times-Roman");

  if(offset!=0.0)
    cout << offset << " " << offset << " translate\n";

  // draw time
  if(drawtime)
    {
      cout << "gsave\n"
	   << "0.25 setlinewidth\n"
	   << "0.5 setgray\n"
	;
      const int xw=static_cast<int>(pt2cm(width))/2;
      for(int x=0; x<xw; ++x)
	{
	  const double X=width*x/xw;

	  // omit first line, since it will be covered by the box
	  if(x)
	    cout << X << " 0 moveto " << X << " " << height << " lineto stroke\n";

	  const int samplenr=static_cast<int>(static_cast<double>(x)/xw*(size/channels));
	  const int min=samplenr/(samplerate*60);
	  const int sec=samplenr/samplerate;
	  const int rest=samplenr%samplerate;
	  const int hun=static_cast<int>(100.0*rest/samplerate);
	  ostringstream os;
	  if(min)
	    os << min << ":";
	  os << sec;
	  if(hun)
	    {
	      os << ".";
	      if(hun<10)
		os << "0";
	      os << hun;
	    }
	  os << "s";
	  print(os.str(), X+0.25*fontsize, 0.25*fontsize);
	}
      cout << "grestore\n";
    }

  // draw samples
  for(int c=0; c<channels; ++c)
    {
#define X(x) x + strokesize
#define Y(x) x*amp+amp + 2.0*amp*c + strokesize
      cout << "newpath\n";
      unsigned i=c;
      double acc=0;
      std::string lastx;
      for(int y=0; y!=scale; ++y, i+=channels)
	acc+=buf[i];
      acc/=scale;
      cout << X(0) << " " << Y(acc) << " moveto\n";
      for(int x=0; x!=frames; ++x)
	{
	  acc=0;
	  for(int y=0; y!=scale; ++y, i+=channels)
	    acc+=buf[i];
	  acc/=scale;

	  // write X position onto text stream
	  ostringstream os;
	  os << fixed << setprecision(2) << X(step*x);
	  // get X position as string
	  std::string x_=os.str();
	  // only output new line when X position has changed
	  if(x_ != lastx)
	    {
	      cout << x_ << " " << Y(acc) << " lineto\n";
	      lastx=x_;
	    }
	}
      cout << strokesize << " setlinewidth stroke\n";

      if(drawbox)
	{
	  // draw null line
	  cout << "newpath 0 " << Y(0) << " moveto " << width << " " << Y(0) << " lineto 0.25 setlinewidth stroke\n";

	  // draw line between channels
	  if(drawbox && channels>1 && c<=channels-1)
	    cout << "newpath 0 " << Y(-1) << " moveto " << width << " " << Y(-1) << " lineto 0.5 setlinewidth stroke\n";
	}
#undef X
#undef Y
    }

  if(text.size())
    {
      string t;
      if(text==" ")
	{
	  ostringstream os;
	  os << filename << " " << samplerate << "Hz " << static_cast<double>(size/channels)/samplerate << "s";
	  t=os.str();
	}
      else
	t=text;

      print(t, 0.25*fontsize, 0.25*fontsize + (drawtime?fontsize:0));
    }

  // draw box
  if(drawbox)
    setcol(0.0) << "fr 1 setlinewidth stroke\n";

  cout << "grestore\n";
  cout << "showpage\n";
  return 0;
}
